// ======================================================================
//
// DetailAppear.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "DetailAppear.h"

#include "Camera.h"
#include "Extents.h"
#include "Gl.h"
#include "Iff.h"
#include "Material.h"

// ======================================================================

const Tag TAG_SUPR = TAG(S,U,P,R);
const Tag TAG_LTAB = TAG(L,T,A,B);
const Tag TAG_FAR  = TAG3(F,A,R);

// ======================================================================

bool DetailAppear::superEnabled(true);
bool DetailAppear::farEnabled(true);
int  DetailAppear::globalForcedLevel(-1);
int  DetailAppear::globalMaxLevel(0);

// ======================================================================

inline const real *DetailAppear::getRangeSquaredTable(void) const
{
	return ((superRangeSquaredTable && superEnabled) ? superRangeSquaredTable : normalRangeSquaredTable);
}

// ----------------------------------------------------------------------

DetailAppear::DetailAppear(const char *iffName, Iff &iff, real newScale)
: MultiAppear(AppearIdDetail, iffName, NULL, newScale)
{
	int   i, ranges;
	real  range;

	if (iff.seekChunk(TAG_SUPR, Iff::FlagOptional))
	{
		ranges = iff.chunkLength(sizeof(float));
		superRangeSquaredTable = MEM_OWN(iff.read_float(Iff::ReadRest));
		iff.exitChunk();
	}
	else
		superRangeSquaredTable = NULL;

	iff.seekChunk(TAG_RANG);
		ranges = iff.chunkLength(sizeof(float));
		normalRangeSquaredTable = MEM_OWN(iff.read_float(Iff::ReadRest));
	iff.exitChunk();

	// create the level index table
	levelTable = NEW int[ranges];
	if (iff.seekChunk(TAG_LTAB, Iff::FlagOptional))
	{
		DEBUG_FATAL(iff.chunkLength(sizeof(int32)) != ranges, ("DetailAppear:: bad LTAB"));
		for (i = 0; i < ranges; i++)
			levelTable[i] = iff.read_int32();
		iff.exitChunk();
	}
	else
	{
		int off;

		// build a workable level table.  super detail level is 0, everything else starts at 1
		off = (superRangeSquaredTable != NULL) ? 1 : 0;
		for (i = 0; i < ranges; i++)
			levelTable[i] = i + off;
	}

	// load in the detail levels
	iff.seekForm(TAG_MESH);
		load(iffName, iff, newScale);
	iff.exitForm();

	DEBUG_FATAL(ranges != maxFrame, ("DetailAppear::bad counts %d %d\n", ranges, maxFrame));

	// if the range table is completely empty, fill it with icky values
	if (superRangeSquaredTable && superRangeSquaredTable[maxFrame-1] == CONST_REAL(0))
	{
		WARNING(true, ("DetailAppear:: empty super range table for %s\n", iffName));
		for (range = CONST_REAL(20), i = 0; i < maxFrame; i++, range *= CONST_REAL(2))
			superRangeSquaredTable[i] = range;
	}

	if (normalRangeSquaredTable[maxFrame-1] == CONST_REAL(0))
	{
		WARNING(true, ("DetailAppear:: empty range table for %s\n", iffName));
		for (range = CONST_REAL(20), i = 0; i < maxFrame; i++, range *= CONST_REAL(2))
			normalRangeSquaredTable[i] = range;
	}

	// square all the ranges for faster finding of the correct detail level later
	if (superRangeSquaredTable)
		for (i = 0; i < maxFrame; i++)
			superRangeSquaredTable[i] = sqr(superRangeSquaredTable[i]);
	for (i = 0; i < maxFrame; i++)
		normalRangeSquaredTable[i] = sqr(normalRangeSquaredTable[i]);

	// load in the shared data
	loadShared(iff, iffName);

	// if we don't load any extents, use the first appearance's extent information
	if (!extents && frame[0]->getExtents())
		setExtents(NEW Extents(frame[0]->getExtents()->getSphereCenter(), frame[0]->getExtents()->getSphereRadius()));

	if (iff.seekChunk(TAG_FAR, Iff::FlagOptional))
	{
		int r, g, b;

		r = iff.read_u_int8();
		g = iff.read_u_int8();
		b = iff.read_u_int8();
		farMaterial = MaterialList::fetch(r, g, b);

		farSize = real(iff.read_int8()) * CONST_REAL(0.5);
		farRangeSquared = iff.read_float();

		iff.exitChunk();
	}
	else
	{
		farMaterial = NULL;
		farSize = 0;
		farRangeSquared = CONST_REAL(0);
	}
}

// ----------------------------------------------------------------------

DetailAppear::~DetailAppear(void)
{
	MaterialList::release(farMaterial);
	farMaterial = NULL;

	delete [] superRangeSquaredTable;
	superRangeSquaredTable = NULL;

	delete [] normalRangeSquaredTable;
	normalRangeSquaredTable = NULL;

	delete [] levelTable;
	levelTable = NULL;
}

// ----------------------------------------------------------------------

const Appear *DetailAppear::getRangedAppear(real rangeSquared) const
{
	int         i;
	const real *table;

	// handle forced detail levels
	if (globalForcedLevel >= 0)
		return getExplicitDetailLevel(globalForcedLevel);

	if (farEnabled && farMaterial && rangeSquared > farRangeSquared)
		return this;

	// skip all detail levels under the specified max
	for (i = 1; i < maxFrame && levelTable[i] <= globalMaxLevel; i++)
		;

	// skip detail levels that are too close
	table = getRangeSquaredTable();
	for ( ; i < maxFrame && table[i] <= rangeSquared; i++)
		;

	// return the last valid frame
	return frame[i-1];
}

// ----------------------------------------------------------------------

const Appear *DetailAppear::getExplicitDetailLevel(int wantLevel) const
{
	int i;

	for (i = 0; i < maxFrame; i++)
		if (levelTable[i] == wantLevel)
			return frame[i];

	return NULL;
}

// ----------------------------------------------------------------------

const Appear *DetailAppear::getRenderAppear(Camera &camera, const Object *object) const
{
	const Appear *appear;

	appear = getRangedAppear(camera.distanceSquaredTo(object));

	// pass the call along to the proper appearance at that range
	if (appear == this)
	{
		Vector v;

		object->getWorldPosition(&v);
		if (!camera.testSphereVisibility(v, CONST_REAL(0)))
			appear = NULL;
	}
	else
		if (appear)
			appear = appear->getRenderAppear(camera, object);
	
	return appear;
}

// ----------------------------------------------------------------------

void DetailAppear::render(Camera &camera, const Xform &objectToWorld) const
{
	Vector v;
	real   x, y, ooz;

	// handle far detail level
	objectToWorld.getPosition(&v);
	camera.rotateTranslateW2O(v, &v);

	if (camera.project(v, &x, &y, Camera::CameraSpace))
	{
		TVertex pv[4];

		ooz = RECIP(v.z);
		for (int i = 0; i < 4; i++)
		{
			pv[i].z   = v.z;
			pv[i].ooz = ooz;
			pv[i].u   = CONST_REAL(0);
			pv[i].v   = CONST_REAL(0);
			pv[i].r   = CONST_REAL(1);
			pv[i].g   = CONST_REAL(1);
			pv[i].b   = CONST_REAL(1);
			pv[i].a   = CONST_REAL(0);
		}

		pv[0].x = x - farSize;
		pv[0].y = y - farSize;
		pv[1].x = x + farSize;
		pv[1].y = y - farSize;
		pv[2].x = x + farSize;
		pv[2].y = y + farSize;
		pv[3].x = x - farSize;
		pv[3].y = y + farSize;

		Gl::setMaterial(farMaterial->getStaticMaterial());
		Gl::drawTriangleFan(pv, 4);
	}
}

// ======================================================================
